"""
Wisconsin Autonomous - https://www.wisconsinautonomous.org

Copyright (c) 2021 wisconsinautonomous.org
All rights reserved.

Use of this source code is governed by a BSD-style license that can be found
in the LICENSE file at the top level of the repo
"""

# ----------------------
# Data loading utilities
# ----------------------

import pathlib
import contextlib

# Grab the data folder in the root of this repo
DATA_DIRECTORY = str(pathlib.Path(__file__).absolute().parent / "data")


def get_wa_data_file(filename):
    """Get the absolute path to the file passed

    Args:
            filename (str): file relative to the data folder to get the absolute path for

    Returns:
            str: the absolute path of the file
    """
    return str(pathlib.Path(DATA_DIRECTORY) / filename)


def set_wa_data_directory(path):
    """Set the data path

    Args:
        path (str): relative (or absolute) path where the data is stored
    """
    global DATA_DIRECTORY

    DATA_DIRECTORY = str(pathlib.Path(path).resolve())


@contextlib.contextmanager
def set_wa_data_directory_temp(path):
    """Set the data path and yield the result. Will restore old directory immediately after

    Args:
        path (str): relative (or absolute) path where the data is stored
    """
    global DATA_DIRECTORY

    OLD_DATA_DIRECTORY = f'{DATA_DIRECTORY}'
    DATA_DIRECTORY = str(pathlib.Path(path).resolve())
    yield
    DATA_DIRECTORY = OLD_DATA_DIRECTORY

# -----------------------
# Type checking utilities
# -----------------------


def check_type(obj, correct_type, variable_name, function_name):
    """Check the type of an object to verify it is the correct type

    Args:
        obj (Any): The original object to check the type of
        correct_type (Any): The type obj should be
        variable_name (str): The name of the original variable for printing
        function_name (str): The name of the calling function
    """
    if not isinstance(obj, correct_type):
        raise TypeError(
            f"{function_name}: '{variable_name}' should be a '{correct_type}' but is '{type(obj)}.")

# ----------------------
# JSON related utilities
# ----------------------


def load_json(filename: str):
    """Load a json file

    Args:
        filename (str): The file to load
    """
    import json

    with open(filename) as f:
        j = json.load(f)

    return j


def check_field(j: dict, field: str, value=None, field_type=None, allowed_values: list = None, optional: bool = False):
    """Check a field in a dict loaded from json

    Args:
        j (dict): The dictionary loaded via a json file
        field (str): The field to check
        value (Any, optional): Some value field must be
        field_type (Type, optional): The type the field must be
        allowed_values: The allowed values
        optional (bool, optional): Whether the field is optional
    """

    if field not in j:
        if optional:
            return

        raise KeyError(
            f"check_field: '{field}' is not in the passed json")

    if value is not None and j[field] != value:
        raise ValueError(
            f"check_field: was expecting '{value}' for '{field}', but got '{j[field]}'.")

    if field_type is not None and not isinstance(j[field], field_type):
        raise TypeError(
            f"check_field: was expecting '{field}' to be '{field_type}', but was '{type(j[field])}'.")

    if allowed_values is not None:
        check_field_allowed_values(j, field, allowed_values)


def check_field_allowed_values(j: dict, field: str, allowed_values: list):
    """Check a field in a dict loaded from json and check that all elements are allowed

    Args:
        j (dict): The dictionary loaded via a json file
        field (str): The field to check
        allowed_values: The allowed values
    """

    check_field(j, field)

    # Make sure each value is present in the allowed_values list
    # Ignore dicts cause those are objects, allow parsing of those separately
    if j[field] not in allowed_values and j[field] is not dict:
        raise ValueError(
            f"check_field_allowed_values: '{j[field]}' is not an allowed value")



"""
Wisconsin Autonomous - https://www.wisconsinautonomous.org

Copyright (c) 2021 wisconsinautonomous.org
All rights reserved.

Use of this source code is governed by a BSD-style license that can be found
in the LICENSE file at the top level of the repo
"""

import argparse
from pyrr import Vector3, Quaternion
from numbers import Number

# ---------------------------------
# Vector/Quaternion/Math core items
# ---------------------------------


class WAVector(Vector3):
    def __new__(cls, value=None, dtype=None):
        if dtype is not None:
            return super().__new__(cls, value, dtype)
        else:
            return super().__new__(cls, value, float)


class WAQuaternion(Quaternion):
    def __new__(cls, value=None, dtype=None):
        if dtype is not None:
            return super().__new__(cls, value, dtype)
        else:
            return super().__new__(cls, value, float)

    def __add__(self, other):
        """Adds two quaternions or a constant"""
        if isinstance(other, WAQuaternion):
            return WAQuaternion([self.x + other.x, self.y + other.y, self.z + other.z, self.w + other.w])
        elif isinstance(other, Number):
            return WAQuaternion([self.x + other, self.y + other, self.z + other, self.w + other])

    def __sub__(self, other):
        """Subs two quaternions or a constant"""
        if isinstance(other, WAQuaternion):
            return WAQuaternion([self.x - other.x, self.y - other.y, self.z - other.z, self.w - other.w])
        elif isinstance(other, Number):
            return WAQuaternion([self.x - other, self.y - other, self.z - other, self.w - other])

    def __mul__(self, other):
        """Multiplies by a constant"""
        if isinstance(other, Number):
            return WAQuaternion([self.x * other, self.y * other, self.z * other, self.w * other])
        else:
            return super().__mul__(other)


# --------------
# CLI core items
# --------------


class WAArgumentParser(argparse.ArgumentParser):
    """Argument parser wrapper.

    Has handy default arguments method

    Args:
        use_sim_defaults (bool, optional): Use the default arguments (can add more). Defaults to True.
        skip_defaults (list, optional): The default arguments to skip.
        description (str, optional): Description used to describe the simulation. Printed by the help menu. Defaults to ("Wisconsin Autonomous Simulator").
    """

    def __init__(self, use_sim_defaults: bool = True, skip_defaults: list = [], description=("Wisconsin Autonomous Simulator")):
        super().__init__(description=description)

        if use_sim_defaults:
            self.add_default_sim_arguments(skip_defaults)

    def add_default_sim_arguments(self, skip_defaults=[]):
        """Adds default arguments for most simulations

        Args:
            skip_defaults (list, optional): The default arguments to skip.

        Includes:
            -s,--sim_step_size: Simulation step size
            -rs,--render_step_size: Rendering step size
            -iv,--irrlicht: Use irrlicht visualization
            -mv,--matplotlib: Use matplotlib visualization
            -r,--record: Record the simulation to a csv
        """
        if 's' not in skip_defaults:
            self.add_argument(
                "-s",
                "--sim_step_size",
                type=float,
                help="Simulation Step Size",
                default=3e-3,
                dest="step_size",
            )

        if 'rs' not in skip_defaults:
            self.add_argument(
                "-rs",
                "--render_step_size",
                type=float,
                help="Render Update Rate [Hz]",
                default=1 / 10.0,
            )

        if 'iv' not in skip_defaults:
            self.add_argument(
                "-iv",
                "--irrlicht",
                action="store_true",
                help="Use Irrlicht to Visualize",
                default=False,
            )

        if 'mv' not in skip_defaults:
            self.add_argument(
                "-mv",
                "--matplotlib",
                action="store_true",
                help="Use Matplotlib to Visualize",
                default=False,
            )

        if 'r' not in skip_defaults:
            self.add_argument(
                "-r",
                "--record",
                action="store_true",
                help="Record Simple State Data",
                default=False,
            )
